-----------------------------------------------------------------
-- Title: AES encryption module with a 128-bit data path --------
-- Filename: aes_enc128.vhd                                    --
-----------------------------------------------------------------
-- Ver 1.0                                                     --
-- January 2017                                                --
-----------------------------------------------------------------
-- Author:                                                     --
--   Fischer V., (C) 2017 Laboratoire Hubert Curien, France    --
-----------------------------------------------------------------
-- Description :                                               --
--                                                             --
-- Implementation   of  the  AES   cipher   with  configurable -- 
-- encryption.                                                 --
-- Sboxes are implemented in a dual port synchronous embedded  --
-- memory organized in 8 x 512 bits, so that two substitutions --
-- can be realized using one memory block.                     --
-- Implementation  is  suitable  for  ACTEL as  well as ALTERA --
-- family (Xilinx should follow).                              --
-----------------------------------------------------------------
--                                                             --
-- Parameters/Inputs/outputs ---------------------             --
-- Parameters:                                                 --
--    * FAMILY : string constant containing name of the family --
--                 Possible options are: "ALTERA", "ACTEL"     --
--                                                             --
-- Inputs :                                                    --
--    * clk        : system clock                              --
--    * reset      : system reset, active high                 --
--    * start      : starts encryption                         --
--    * data_in    : plaintext (128 bits)                      --
--    * key_in     : key input (128 bits)                      --
--                                                             --
-- Outputs :                                                   --
--    * busy       : enciphering flag                          --
--    * data_out   : ciphertext (128 bits)                     --
--                                                             --
-----------------------------------------------------------------
-- Revision history:                                           --
--                                                             --
-----------------------------------------------------------------
LIBRARY ieee; 
USE ieee.std_logic_unsigned.all;
USE ieee.std_logic_1164.all;

LIBRARY work;
USE work.aes_pkg.ALL;

ENTITY aes_enc128 IS
  GENERIC(
	  FAMILY   : STRING  := FAMILY_VER
    );
  PORT (
    clk         : IN STD_LOGIC;
    reset       : IN STD_LOGIC;
--    start       : IN STD_LOGIC;
    ser_in      : IN STD_LOGIC;
--    key_in      : IN STD_LOGIC_VECTOR (127 DOWNTO 0);
    busy        : OUT STD_LOGIC;
    data_out    : OUT STD_LOGIC_VECTOR (127 DOWNTO 0)
    );
END aes_enc128;

-------------- A R C H I T E C T U R E ------------------

ARCHITECTURE rtl OF aes_enc128 IS

-- MixColums block
  COMPONENT mixcols IS
    PORT (
      data_in       : IN STD_LOGIC_VECTOR (31 DOWNTO 0) := X"00000000";
      data_out      : OUT STD_LOGIC_VECTOR (31 DOWNTO 0)
      );
  END COMPONENT;

-- Substitution block using true dual-port memory 
-- (Actel: with sbox generator)
  COMPONENT sbox_wrap IS
    GENERIC (
      SBOX_COUNT : integer := 16;
      FAMILY     : string := "ACTEL"
      );
    PORT (
      clk         : IN STD_LOGIC;
      reset       : IN STD_LOGIC; 
      sbox_in     : IN STD_LOGIC_VECTOR (SBOX_COUNT*8-1 DOWNTO 0);
      sbox_out    : OUT STD_LOGIC_VECTOR (SBOX_COUNT*8-1 DOWNTO 0);
      initdone    : OUT STD_LOGIC
      );
  END COMPONENT;
 
-- Key scheduler for encryption and decryption (yet 128bit operation only)
  COMPONENT keysched IS
    GENERIC(
      FAMILY    : string := "ACTEL"
      );
    PORT(
      clk          : IN STD_LOGIC;
      reset        : IN STD_LOGIC;  -- global reset
      key_exp_en   : IN STD_LOGIC;  -- key generation enabled
      key_in       : IN STD_LOGIC_VECTOR(127 DOWNTO 0);
      initdone     : OUT STD_LOGIC;
      key_out      : OUT STD_LOGIC_VECTOR(127 DOWNTO 0)
      );
  END COMPONENT;
  
-- Select family
  -- IMPORTANT!!! : when using altera, please uncomment
  --                usage of LIBRARY altera_mf, and USE altera_mf.all
  --                in file tdp_mem8x512.vhd!!!! 
  
  -- IMPORTANT!!! : when using actel, please uncomment
  --                usage of LIBRARY fusion ans USE fusion.all
  --                in file dp_mem8x512.vhd !!!

-- *** CONTROL SIGNALS ***

  -- Main state machine
  TYPE SM_ENC_TYPE IS (sbox_init, enc_waiting, enc_running);
  SIGNAL sm_enc: SM_ENC_TYPE;

  -- Round counter
  SIGNAL round  : integer RANGE 0 TO 12;

  SIGNAL initdone      : STD_LOGIC; -- main sboxes initialization done
  SIGNAL ksch_initdone : STD_LOGIC; -- key scheduler sboxes initialization done

  SIGNAL finished      : STD_LOGIC := '0'; -- finished serial to parallel conv.
  SIGNAL start         : STD_LOGIC := '0'; -- start of encryption (ser2par full)

  SIGNAL ser2par       : STD_LOGIC_VECTOR (255 DOWNTO 0) := (OTHERS => '0');
  
  SIGNAL data_in       : STD_LOGIC_VECTOR (127 DOWNTO 0) := (OTHERS => '0'); 
  SIGNAL key_in        : STD_LOGIC_VECTOR (127 DOWNTO 0) := (OTHERS => '0'); 
  
  SIGNAL cnt_in        : STD_LOGIC_VECTOR (7 DOWNTO 0) := (OTHERS => '0');
 
--  SIGNAL delay_start   : STD_LOGIC; -- start delayed by 1 clock
  SIGNAL key_exp_en    : STD_LOGIC; -- key expansion enabled

-- Input multiplexer for round key addition to input data or output data
  SIGNAL mux_in        : STD_LOGIC_VECTOR (127 DOWNTO 0); 
  
-- Round key addition
  SIGNAL sbst_in       : STD_LOGIC_VECTOR (127 DOWNTO 0); 
  SIGNAL add_rkey_last : STD_LOGIC_VECTOR (127 DOWNTO 0); 

-- Substitution result
  SIGNAL sbst_out   : STD_LOGIC_VECTOR (127 DOWNTO 0); 
  SIGNAL sbst_out0  : STD_LOGIC_VECTOR (31 DOWNTO 0);      
  SIGNAL sbst_out1  : STD_LOGIC_VECTOR (31 DOWNTO 0);      
  SIGNAL sbst_out2  : STD_LOGIC_VECTOR (31 DOWNTO 0);      
  SIGNAL sbst_out3  : STD_LOGIC_VECTOR (31 DOWNTO 0);      

-- Byte rotation result
  SIGNAL shft_row  : STD_LOGIC_VECTOR (127 DOWNTO 0);
  --
  SIGNAL shft_row0 : STD_LOGIC_VECTOR (31 DOWNTO 0);
  SIGNAL shft_row1 : STD_LOGIC_VECTOR (31 DOWNTO 0);
  SIGNAL shft_row2 : STD_LOGIC_VECTOR (31 DOWNTO 0);
  SIGNAL shft_row3 : STD_LOGIC_VECTOR (31 DOWNTO 0);

-- Multiplexing between inverse and direct Byte rotation result
  SIGNAL mux_shft_row  : STD_LOGIC_VECTOR (127 DOWNTO 0);

-- MixColumns result
  SIGNAL mixclmns  : STD_LOGIC_VECTOR (127 DOWNTO 0); -- For simulation only
  SIGNAL mixclmns0 : STD_LOGIC_VECTOR (31 DOWNTO 0);
  SIGNAL mixclmns1 : STD_LOGIC_VECTOR (31 DOWNTO 0);
  SIGNAL mixclmns2 : STD_LOGIC_VECTOR (31 DOWNTO 0);
  SIGNAL mixclmns3 : STD_LOGIC_VECTOR (31 DOWNTO 0);

-- Round key signals
  SIGNAL roundkey    : STD_LOGIC_VECTOR (127 DOWNTO 0); -- For simulation only

BEGIN

-- Family select messages
  ASSERT (FAMILY = "ALTERA" OR FAMILY = "ACTEL" OR FAMILY = "XILINX")
  REPORT "FPGA family must be ALTERA, ACTEL or XILINX!"
  SEVERITY ERROR;   -- Stop compilation!!!

  ASSERT (FAMILY /= "ALTERA")
  REPORT "Implementing the design in Altera FPGA family!"
  SEVERITY WARNING; -- Just send a warning

  ASSERT (FAMILY /= "ACTEL")
  REPORT "Implementing the design in Actel FPGA family!"
  SEVERITY WARNING; -- Just send a warning

-------------------------------------------------------------
-- Input serial to parallel converter
-------------------------------------------------------------
  PROCESS(clk, reset)
  BEGIN
    IF reset = '1' THEN
      ser2par <= (OTHERS => '0');
    ELSIF rising_edge(clk) THEN
      ser2par(255) <= ser_in;
      ser2par(254 DOWNTO 0) <= ser2par(255 DOWNTO 1);
    END IF;
  END PROCESS;

-------------------------------------------------------------
-- Counter of input bits
-------------------------------------------------------------
  PROCESS(clk, reset)
  BEGIN
    IF reset = '1' THEN
      cnt_in <= (OTHERS => '0');
    ELSIF rising_edge(clk) THEN
      cnt_in <= cnt_in + 1;
      IF cnt_in = 255 THEN
        finished <= '1';
      ELSE
        finished <= '0';
      END IF;
      IF finished = '1' THEN
        key_in  <= ser2par(255 DOWNTO 128);
        data_in <= ser2par(127 DOWNTO 0);
        start   <= '1';
      ELSE 
        start   <= '0';
      END IF;
    END IF;
  END PROCESS;
  
  
-----------------------
------ DATA PATH ------
-----------------------

  -- Input multiplexer for encryption
  mux_in <= data_in WHEN (round = 0) ELSE mixclmns;
    
  -- Addition of the round key
  sbst_in <= roundkey XOR mux_in;

-- Substitution operation using ROMs (family independent)
  sbst: sbox_wrap
    GENERIC MAP(
      SBOX_COUNT => 16,
      FAMILY  => FAMILY
      )
    PORT MAP(
      clk        => clk,
      reset      => reset,
      sbox_in    => sbst_in,
      sbox_out   => sbst_out,
      initdone   => initdone
      );

-- Preparation of shift_rows
  sbst_out3 <= sbst_out(127 DOWNTO 96);
  sbst_out2 <= sbst_out(95 DOWNTO 64);
  sbst_out1 <= sbst_out(63 DOWNTO 32);
  sbst_out0 <= sbst_out(31 DOWNTO 0);

-- Shift rows operation
  shft_row3(31 DOWNTO 24) <= sbst_out3(31 DOWNTO 24);
  shft_row3(23 DOWNTO 16) <= sbst_out2(23 DOWNTO 16);
  shft_row3(15 DOWNTO 8)  <= sbst_out1(15 DOWNTO 8);
  shft_row3(7 DOWNTO 0)   <= sbst_out0(7 DOWNTO 0);
  --
  shft_row2(31 DOWNTO 24) <= sbst_out2(31 DOWNTO 24);
  shft_row2(23 DOWNTO 16) <= sbst_out1(23 DOWNTO 16);
  shft_row2(15 DOWNTO 8)  <= sbst_out0(15 DOWNTO 8);
  shft_row2(7 DOWNTO 0)   <= sbst_out3(7 DOWNTO 0);
  --
  shft_row1(31 DOWNTO 24) <= sbst_out1(31 DOWNTO 24);
  shft_row1(23 DOWNTO 16) <= sbst_out0(23 DOWNTO 16);
  shft_row1(15 DOWNTO 8)  <= sbst_out3(15 DOWNTO 8);
  shft_row1(7 DOWNTO 0)   <= sbst_out2(7 DOWNTO 0);
  --
  shft_row0(31 DOWNTO 24) <= sbst_out0(31 DOWNTO 24);
  shft_row0(23 DOWNTO 16) <= sbst_out3(23 DOWNTO 16);
  shft_row0(15 DOWNTO 8)  <= sbst_out2(15 DOWNTO 8);
  shft_row0(7 DOWNTO 0)   <= sbst_out1(7 DOWNTO 0);
 
-- MixColumn operation
  mixcol3: mixcols PORT MAP (shft_row3, mixclmns3);
  mixcol2: mixcols PORT MAP (shft_row2, mixclmns2);
  mixcol1: mixcols PORT MAP (shft_row1, mixclmns1);
  mixcol0: mixcols PORT MAP (shft_row0, mixclmns0);
  --  
  mixclmns      <= mixclmns3 & mixclmns2 & mixclmns1 & mixclmns0;
 
  -- Addition of the last round key (without MixColumn)
  shft_row      <= shft_row3 & shft_row2 & shft_row1 & shft_row0;
  add_rkey_last <= shft_row XOR roundkey;
  
-- Key scheduling
  keysched_0: keysched
    GENERIC MAP(
      FAMILY       => FAMILY
      )
    PORT MAP(
      clk        => clk,
      reset      => reset,
      key_exp_en => key_exp_en,
      key_in     => key_in,
      initdone   => ksch_initdone,
      key_out    => roundkey
      );

-- Output process
  PROCESS(clk, reset)
  BEGIN
    IF reset = '1' THEN
      data_out <= (OTHERS => '0');
    ELSIF rising_edge(clk) THEN
      IF ((round = 10) AND (sm_enc = enc_running)) THEN 
        data_out <= add_rkey_last;                  
      END IF;
    END IF;
  END PROCESS;

-----------------------
---- CONTROL LOGIC ----
-----------------------

-- Output signal 'Busy' definition
  WITH sm_enc SELECT
    busy   <= '1'  WHEN  sbox_init, -- For Microsemi only
              '1'  WHEN  enc_running,
              '0'  WHEN  OTHERS;

-- Enabling key expansion
  WITH sm_enc SELECT
    key_exp_en <= '1'  WHEN  enc_running,
                  '0'  WHEN  OTHERS;

-- Main state machine
  PROCESS (clk, reset)
  BEGIN 
    IF reset = '1' THEN
      sm_enc <= sbox_init;
    ELSIF rising_edge(clk) THEN
      CASE sm_enc IS
        WHEN sbox_init =>
          IF (initdone AND ksch_initdone) = '1' THEN
            sm_enc <= enc_waiting;
          ELSE
            sm_enc <= sbox_init;
          END IF;
        WHEN enc_waiting =>
          IF start = '1' THEN
            sm_enc <= enc_running;
          END IF;
        WHEN enc_running =>
          IF (round = 10) THEN
            sm_enc <= enc_waiting;
          END IF;
      END CASE;
    END IF;
  END PROCESS;

-- Round counter
  PROCESS(clk)
  BEGIN
    IF rising_edge(clk) THEN
      IF ((start = '1') OR (round = 10)) THEN
        round <= 0;
      ELSE
        round <= round + 1;
      END IF;
    END IF;
  END PROCESS;


END rtl;