-----------------------------------------------------------------
-- Title: Key  scheduler  for  encryption and  decryption      --  
-- Filename: keysched.vhd                                      --
-----------------------------------------------------------------
-- Ver 1.0                                                     --
-- January 26th, 2017                                          --
-----------------------------------------------------------------
-- Author:                                                     --
--   Fischer V., (C) 2017 Laboratoire Hubert Curien, France    --
-----------------------------------------------------------------
-- Description :                                               --
--                                                             --
-- Key scheduler originate from the first key. Next round keys -- 
-- are computed each per one clock cycle. Scheduler is able to --
-- compute keys in both directions (forward for encryption and --
-- backward for decryption).                                   --
--                                                             --
-----------------------------------------------------------------
-- Parameters:                                                 --
--   * FAMILY   : string constant containing name of the family--
--                 Possible options are: "ALTERA", "ACTEL"     --
-- Inputs :                                                    --
--    * clk       : system clock                               -- 
--    * reset    : global reset, active high                   -- 
--    * key_in    : 128-bit first round key (original key)     -- 
--    * key_exp_en: HIGH enable  round keys to be generated    -- 
--                                                             --
-- Outputs :                                                   --
--    * initdone   : Sbox RAM initialization done flag         --
--    * key_out    : 128-bit round key output                  --
-----------------------------------------------------------------
-- Revision history:                                           --
--                                                             --
-----------------------------------------------------------------
LIBRARY ieee;
USE ieee.std_logic_1164.all;
USE ieee.std_logic_unsigned.all; 


ENTITY keysched IS
  GENERIC(
    FAMILY   : string := "ACTEL" 
    );
  PORT(
    clk          : IN std_logic;
    reset        : IN std_logic;  
    key_exp_en   : IN std_logic;  
    key_in       : IN std_logic_vector(127 DOWNTO 0);
    initdone     : OUT std_logic;
    key_out      : OUT std_logic_vector(127 DOWNTO 0)
    );
END keysched;

-------------- A R C H I T E C T U R E ------------------

ARCHITECTURE behavioral OF keysched IS
  
  COMPONENT sbox_wrap IS
    GENERIC (
      SBOX_COUNT : integer := 16;
      FAMILY     : string := "ACTEL"
      );
    PORT (
      clk         : IN std_logic;
      reset       : IN std_logic; 
      sbox_in     : IN std_logic_vector (SBOX_COUNT*8-1 DOWNTO 0);
      sbox_out    : OUT std_logic_vector (SBOX_COUNT*8-1 DOWNTO 0);
      initdone    : OUT std_logic
      );
  END COMPONENT;

  
-- main datapath signals 
  SIGNAL w0_in : std_logic_vector(31 DOWNTO 0);
  SIGNAL w1_in : std_logic_vector(31 DOWNTO 0);
  SIGNAL w2_in : std_logic_vector(31 DOWNTO 0);
  SIGNAL w3_in : std_logic_vector(31 DOWNTO 0); 
  SIGNAL w0_out : std_logic_vector(31 DOWNTO 0);
  SIGNAL w1_out : std_logic_vector(31 DOWNTO 0);
  SIGNAL w2_out : std_logic_vector(31 DOWNTO 0);
  SIGNAL w3_out : std_logic_vector(31 DOWNTO 0);
  
  -- multiplexing between enc. and decryption 
  SIGNAL xmux0_out : std_logic_vector(31 DOWNTO 0);
  SIGNAL xmux1_out : std_logic_vector(31 DOWNTO 0);
  SIGNAL xmux2_out : std_logic_vector(31 DOWNTO 0);
  SIGNAL xmuxf_out : std_logic_vector(31 DOWNTO 0);
  
  -- generated keys one clock ahead when compared to output keys
  SIGNAL w4  : std_logic_vector(31 DOWNTO 0);
  SIGNAL w5  : std_logic_vector(31 DOWNTO 0);
  SIGNAL w6  : std_logic_vector(31 DOWNTO 0);
  SIGNAL w7  : std_logic_vector(31 DOWNTO 0);
  
  -- generated key part w11 is:
  -- ENCRYPTION: one clock ahead when compared to w3 output key
  -- DECRYPTION: two clocks ahead when compared to w3 output key
  SIGNAL w11 : std_logic_vector(31 DOWNTO 0);

   -- for implementation of function 
  SIGNAL w3_rot : std_logic_vector(31 DOWNTO 0);
  SIGNAL sb_out : std_logic_vector(31 DOWNTO 0);
  SIGNAL func_out : std_logic_vector(31 DOWNTO 0);
  
  -- RCON and inverse RCON generation
  SIGNAL rcon_d : std_logic_vector(7 DOWNTO 0);
  SIGNAL rcon_q : std_logic_vector(7 DOWNTO 0);
  SIGNAL inv_rcon_d : std_logic_vector(7 DOWNTO 0);
  SIGNAL inv_rcon_q : std_logic_vector(7 DOWNTO 0);
  SIGNAL inv_rcon_q_tmp : std_logic_vector(7 DOWNTO 0);
  SIGNAL rconmux_out : std_logic_vector(7 DOWNTO 0);
  
 -- main control signals  
  SIGNAL mux_sel  : std_logic; 
  SIGNAL rcon_nrst : std_logic;
    
  SIGNAL sb_init_done : std_logic;

------------------ B E G I N ------------------  
BEGIN
    
-------------------------------
---------- DATA PATH ---------- 
-------------------------------

-- Round key temporary storage
  Wreg: PROCESS (reset, clk)
  BEGIN
    IF(reset = '1') THEN
      w0_out <= (OTHERS => '0');
      w1_out <= (OTHERS => '0');
      w2_out <= (OTHERS => '0');
      w3_out <= (OTHERS => '0');
    ELSIF rising_edge(clk) THEN
        w0_out <= w0_in;
        w1_out <= w1_in;
        w2_out <= w2_in;
        w3_out <= w3_in;
    END IF;
  END PROCESS;

-- Implementation of 128 multipexers for xors
xmux0_out <= w4;
xmux1_out <= w5;
xmux2_out <= w6;

-- Implementation of 32 xors and 32 multiplexers to precompute input to Function
w11 <= w2_in XOR w3_in;
xmuxf_out <= w3_in;

-- Implementation of 128 xors 
w4 <= w0_out XOR func_out;
w5 <= w1_out XOR xmux0_out;
w6 <= w2_out XOR xmux1_out;
w7 <= w3_out XOR xmux2_out;

-- 128 multiplexers(for switching between feedback and loading the original key)
w0_in <= w4 WHEN mux_sel = '1' ELSE key_in(127 DOWNTO 96);
w1_in <= w5 WHEN mux_sel = '1' ELSE key_in(95 DOWNTO 64);
w2_in <= w6 WHEN mux_sel = '1' ELSE key_in(63 DOWNTO 32);
w3_in <= w7 WHEN mux_sel = '1' ELSE key_in(31 DOWNTO 0);

-- Key output
  key_out(127 DOWNTO 96) <= w0_out;
  key_out(95 DOWNTO 64)  <= w1_out;
  key_out(63 DOWNTO 32)  <= w2_out;
  key_out(31 DOWNTO 0)   <= w3_out;


-- FUNCTION IMPLEMENTATION
  -- Rotation of the 32-bit word
    w3_rot(31 DOWNTO 24) <= xmuxf_out(23 DOWNTO 16);
    w3_rot(23 DOWNTO 16) <= xmuxf_out(15 DOWNTO  8);
    w3_rot(15 DOWNTO  8) <= xmuxf_out(7  DOWNTO  0);
    w3_rot(7  DOWNTO  0) <= xmuxf_out(31 DOWNTO 24);
  
  -- Instantiation of 4 sboxes (2 blocks of RAM)
    sbst: sbox_wrap
      GENERIC MAP(
        SBOX_COUNT => 4,
        FAMILY     => FAMILY
        )
      PORT MAP(
        clk        => clk,
        reset     => reset,
        sbox_in    => w3_rot,
        sbox_out   => sb_out,
        initdone   => sb_init_done
        );

  -- RCON multiplexer
    rconmux_out <= rcon_q;

  -- XOR with RCON constant
    func_out(31 DOWNTO 24) <= sb_out(31 DOWNTO 24) XOR rconmux_out;
    func_out(23 DOWNTO  0) <= sb_out(23 DOWNTO  0);   


-----------------------------------
---------- CONTROL LOGIC ---------- 
-----------------------------------

-- rams are initialized
  initdone <= sb_init_done ;  

-- RCON gen. enable/reset
  rcon_nrst <= key_exp_en;    

-- first key/next keys
  mux_sel <= key_exp_en;      

-- Computation of RCON constant by multiplying by 2
  PROCESS (rcon_nrst, clk)
  BEGIN
    IF(rcon_nrst = '0') THEN
      rcon_q <= X"01";    
    ELSIF rising_edge(clk) THEN
      rcon_q <= rcon_d; 
    END IF;
  END PROCESS;

  -- Multiplication by 2 in GF(2^8), for AES polynomial
    rcon_d(7) <= rcon_q(6);
    rcon_d(6) <= rcon_q(5);
    rcon_d(5) <= rcon_q(4);
    rcon_d(4) <= rcon_q(3) XOR rcon_q(7);
    rcon_d(3) <= rcon_q(2) XOR rcon_q(7);
    rcon_d(2) <= rcon_q(1);
    rcon_d(1) <= rcon_q(0) XOR rcon_q(7);
    rcon_d(0) <= rcon_q(7);

END ARCHITECTURE behavioral; 