-----------------------------------------------------------------
-- Title: 32-bit MixColumns function implementation            --
-- Filename: mixcols.vhd                                       --
-----------------------------------------------------------------
-- Ver 1.0                                                     --
-- November 2008                                               --
-----------------------------------------------------------------
-- Author:                                                     --
-- Fischer V.                                                  --
-- (C) 2008 Laboratoire Hubert Curien                          --
-- France                                                      --
-----------------------------------------------------------------
-- Description :                                               --
--                                                             --
-- Matrix multiplication in MixColumn operation                --
-- (multiplication by constants):                              --
-- b3 = 2*a3 + 3*a2 +  a1  +  a0 ;                             --
-- b2 =  a3  + 2*a2 + 3*a1 +  a0 ;                             --
-- b1 =  a3  +  a2  + 2*a1 + 3*a0;                             --
-- b0 = 3*a3 +  a2  +  a1  + 2*a0;                             --
--                                                             --
-- This solution is based on coeficients multiplication        --
-- using Xtime function (multiplication by two). So            --
-- the first line (b3) can be expressed as:                    --
-- b3 = 2*a3 + 3*a2 + a1 + a0 =                                --
--    = (a3 + a2 + a1 + a0) + 2*(a3 + a2) - a3                 --
-- what is equivalent to                                       --
-- b3 = (a3 + a2 + a1 + a0) + 2*(a3 + a2) + a3                 --
--                                                             --
-- Inputs :                                                    --
--    * data_in   : 32-bit data input                          --
--                                                             --
-- Outputs :                                                   --
--    * data_out  : 32-bit data output                         --
-----------------------------------------------------------------
LIBRARY ieee;
USE ieee.std_logic_1164.ALL;

ENTITY mixcols IS
  PORT (
    data_in       : IN std_logic_vector (31 DOWNTO 0) := X"00000000";
    data_out      : OUT std_logic_vector (31 DOWNTO 0)
     );
END mixcols;

ARCHITECTURE rtl OF mixcols IS

  COMPONENT xtime IS
    PORT (
      data_in  : IN std_logic_vector (7 DOWNTO 0);
      data_out : OUT std_logic_vector (7 DOWNTO 0)
      );
  END COMPONENT;

  SIGNAL a3, a2, a1, a0                     : std_logic_vector(7 DOWNTO 0);
  SIGNAL xor32, xor21, xor10, xor03         : std_logic_vector(7 DOWNTO 0);
  SIGNAL xor32x2, xor21x2, xor10x2, xor03x2 : std_logic_vector(7 DOWNTO 0);
  SIGNAL xor_all                            : std_logic_vector(7 DOWNTO 0);

BEGIN

  a3 <= data_in(31 DOWNTO 24);
  a2 <= data_in(23 DOWNTO 16);
  a1 <= data_in(15 DOWNTO 8);
  a0 <= data_in(7 DOWNTO 0);

  xor10 <= a1 XOR a0;
  xor21 <= a2 XOR a1;
  xor32 <= a3 XOR a2;
  xor03 <= a0 XOR a3;

-- Addition of four bytes
  xor_all <= a0 XOR a1 XOR a2 XOR a3;

-- Multiplication by two of four couples of bytes
  ixor32x2  : xtime PORT MAP (xor32,xor32x2);    -- (a3+a2) * 2
  ixor21x2  : xtime PORT MAP (xor21,xor21x2);    -- (a2+a1) * 2
  ixor10x2  : xtime PORT MAP (xor10,xor10x2);    -- (a1+a0) * 2
  ixor03x2  : xtime PORT MAP (xor03,xor03x2);    -- (a0+a3) * 2

-- Output signal
  data_out(31 DOWNTO 24) <=  xor_all XOR xor32x2 XOR a3;-- 2*a3 + 3*a2 +  a1  +  a0
  data_out(23 DOWNTO 16) <=  xor_all XOR xor21x2 XOR a2;--  a3  + 2*a2 + 3*a1 +  a0
  data_out(15 DOWNTO 8)  <=  xor_all XOR xor10x2 XOR a1;--  a3  +  a2  + 2*a1 + 3*a0
  data_out(7 DOWNTO 0)   <=  xor_all XOR xor03x2 XOR a0;-- 3*a3 +  a2  +  a1  + 2*a0

END rtl;
