-----------------------------------------------------------------
-- Title: Sbox block based on dual port memories for Actel or  --
--        Altera family                                        --
-- Filename: sbox_wrap.vhd                                     --
-----------------------------------------------------------------
-- Ver 3.0                                                     --
-- January 2017                                                --
-----------------------------------------------------------------
-- Author:                                                     --
--   Fischer V., (C) 2017 Laboratoire Hubert Curien, France    --
-----------------------------------------------------------------
-- Description :                                               --
--                                                             --
-- sbox_wrap entity  wraps sboxes for two families: Altera and --
-- Actel. User  can  select sboxes for  which family to use by --
-- changing  FAMILY  parameter. Second   parameter  SBOX_COUNT --
-- specifies number of used sboxes (number of used memories).  --
-- Every two sboxes occupy one dual port memory block.         --   
--                                                             --
-- There  are   some  differences   between  both  families.   --
-- Main difference is in memory initialization.  While RAMs in --
-- Altera hardware can be initialized during FPGA configuration--
-- in Actel RAMs  have to be initialized  manualy after reset! --
-- Therefore  sbox generator  was required  in case  of Actel. --
--                                                             --
-----------------------------------------------------------------
-- Parameters:                                                 --
--   * SBOX_COUNT : number of sboxes => number of DP memories  --
--                                   => width of signals       --
--   * FAMILY     : string constant containing name of family  --
--                  Possible options are: "ALTERA", "ACTEL"    --
-- Inputs :                                                    --
--    * clk       : system clock                               --
--    * reset     : global reset, active high                  --
--    * sbox_in   : address for parallel sboxes                --
--                                                             --
-- Outputs :                                                   --
--    * initdone   : Sbox RAM initialization done flag         --
--    * sboxa_out  : output data of parallel sboxes            --
-----------------------------------------------------------------
LIBRARY ieee;
USE ieee.std_logic_1164.all;

ENTITY sbox_wrap IS
  GENERIC (
    SBOX_COUNT : integer := 16;
    FAMILY     : string := "ACTEL"
      );
  PORT (
    clk         : IN std_logic;
    reset       : IN std_logic;
    sbox_in     : IN std_logic_vector (SBOX_COUNT*8-1 DOWNTO 0);
    sbox_out    : OUT std_logic_vector (SBOX_COUNT*8-1 DOWNTO 0);
    initdone    : OUT std_logic
     );
END sbox_wrap;


ARCHITECTURE structural OF sbox_wrap IS
  
   ---- ****** A L T E R A ******* ------
  COMPONENT sbox IS
    PORT (
      clk         : IN std_logic;
      sboxa_in    : IN std_logic_vector (7 DOWNTO 0);
      sboxb_in    : IN std_logic_vector (7 DOWNTO 0);
      sboxa_out   : OUT std_logic_vector (7 DOWNTO 0);
      sboxb_out   : OUT std_logic_vector (7 DOWNTO 0)
     );
  END COMPONENT;

------------------ B E G I N ------------------   
BEGIN

 ---- ****** A L T E R A ******* ------

 -- When using ALTERA family
 sbox_altera: IF FAMILY = "ALTERA" GENERATE
   -- Generate these sboxes  
   sboxes: FOR i IN 1 TO SBOX_COUNT/2 GENERATE
     sbst: sbox
       PORT MAP (
         clk       => clk, 
         sboxa_in  => sbox_in((2*i-1)*8-1 DOWNTO (2*i-2)*8),
         sboxb_in  => sbox_in((2*i-0)*8-1 DOWNTO (2*i-1)*8),
         sboxa_out => sbox_out((2*i-1)*8-1 DOWNTO (2*i-2)*8),
         sboxb_out => sbox_out((2*i-0)*8-1 DOWNTO (2*i-1)*8)
         );
   END GENERATE sboxes;
   initdone <= '1'; -- no initialization required in altera FPGA
 END GENERATE sbox_altera;

END structural;